#include "Rational2.h"
#include <catch.h>
#include <sstream>

using rational::Rational;

TEST_CASE("Rational_0")
{
    Rational r;
    CHECK(0 == r.numerator());
    CHECK(1 == r.denominator());
}

TEST_CASE("Rational_n")
{
    Rational r(5);
    CHECK(5 == r.numerator());
    CHECK(1 == r.denominator());
}

TEST_CASE("Rational_n_m")
{
    Rational r1(2, 3);
    CHECK(2 == r1.numerator());
    CHECK(3 == r1.denominator());

    Rational r2(4, 6);
    CHECK(2 == r2.numerator());
    CHECK(3 == r2.denominator());

    Rational r3(4, -6);
    CHECK(-2 == r3.numerator());
    CHECK(3 == r3.denominator());

    Rational r4(-4, -6);
    CHECK(2 == r4.numerator());
    CHECK(3 == r4.denominator());
}

TEST_CASE("Rational_den0_throws") {
    CHECK_THROWS_AS(Rational(5, 0), std::overflow_error);
}

TEST_CASE("Rational_equals")
{
    Rational r1(2, 3);
    Rational r2(4, 6);
    CHECK(r1 == r2);
}

TEST_CASE("Rational_comparisons")
{
    Rational r1(3, 10);
    Rational r2(4, 10);

    CHECK(r1 < r2);
    CHECK(r1 <= r2);
    CHECK(r2 > r1);
    CHECK(r2 >= r1);
    CHECK_FALSE(r1 > r2);
    CHECK_FALSE(r1 >= r2);
    CHECK_FALSE(r2 < r1);
    CHECK_FALSE(r2 <= r1);
    CHECK(r1 <= r1);
    CHECK(r2 <= r2);
    CHECK(r1 >= r1);
    CHECK(r2 >= r2);
}

TEST_CASE("Format")
{
    Rational r1(5);
    Rational r2(5, 6);
    std::ostringstream fmt;
    fmt << r1 << ' ' << r2;
    CHECK("5 5/6" == fmt.str());
}

TEST_CASE("To_double")
{
    Rational r1(5);
    Rational r2(1, 3);

    CHECK(5 == to_double(r1));
    CHECK(0.3333 == Approx(to_double(r2)).epsilon(0.0001));
}

TEST_CASE("Negate")
{
    Rational r1(5);
    Rational r2(1, 3);

    CHECK(Rational(-5) == -r1);
    CHECK(Rational(-1, 3) == -r2);
}

TEST_CASE("Reciprocal")
{
    Rational r1(5);
    Rational r2(1, 3);

    CHECK(Rational(1, 5) == r1.reciprocal());
    CHECK(Rational(3) == r2.reciprocal());
}

TEST_CASE("Reciprocal_0_throws")
{
    Rational r(0);
    CHECK_THROWS_AS(r.reciprocal(), std::overflow_error);
}

TEST_CASE("Rational_addition")
{
    using R = Rational;

    CHECK(R(11) == R(5) + R(6));
    CHECK(R(2, 3) == R(1,3) + R(1,3));
    CHECK(R(7, 12) == R(1,3) + R(1,4));
    CHECK(R(1) == R(5,12) + R(7,12));
    CHECK(R(35, 12) == R(2, 3) + R(9, 4));
    CHECK(R(3, 8) == R(1, 8) + R(1, 4));
}

TEST_CASE("Rational_multiplication")
{
    using R = Rational;

    CHECK(R(12) == R(3) * R(4));
    CHECK(R(1) == R(1,3) * R(3));
    CHECK(R(1) == R(2,3) * R(3,2));
}

